#include <tap.h>
#include "libFireBird.h"

TAP_ID( 0x81a0fe99 );
#define TAPNAME "TestRepTimer"
TAP_PROGRAM_NAME(TAPNAME);
TAP_AUTHOR_NAME("R2-D2");
TAP_DESCRIPTION("Adds a repeating timer for the three main types to show bugs");
TAP_ETCINFO(__DATE__);

// Returns 1 if patch added, 2 if already present, and 0 if code cannot be found (unsupported firmware)
int FixTimerAPI()
{
  dword *fw = (dword*)0x80000000;
  if(PatchIsInstalled(fw, "Ta"))
    return 2;
  else
  {
    // TAP_Timer_Modify @ 80002560, copy @+0 to @+5, @+12 ($gp restore), copy @+12 to @+1 (3c0ffffe)
    dword* codeSeg = (void*)FindInstructionSequence(
      "02a0e025 33deffff 3418fffe 53d80005 97999fd4 33deffff 3c19ffff 10000023 03d91021 97989fd6 0019ca00 0338c821 3c0ffffe 1000001d 032f1021",
      "00000000 ffffffff ffffffff ffffffff ffff0000 00000000 ffffffff ffffff00 ffffffff ffff0000 ffffffff ffffffff 00000000 ffffff00 ffffffff",
      0x80170000, 0x80230000, 0, FALSE);
    dword setT7 = 0x3c0ffffe;
    if(!codeSeg)  // Not found, unsupported firmware
      return 0;
    else if(codeSeg[0] == setT7 && codeSeg[5] == codeSeg[12])  // Already patched (no ID?)
      return 2;
    else if(codeSeg[12] == setT7)  // Not patched, so do patching
    {
      codeSeg[5]  = codeSeg[0];
      codeSeg[12] = codeSeg[0];
      codeSeg[0]  = setT7;
      PatchInstallID(fw, "Ta");
      return 1;
    }
    else
      return 0; // Unknown state
  }
}

XDATA char *DaysOfWeek[] = {"Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"};
XDATA char *Months[] = {"???", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

TYPE_TimerInfo info;

char *ResType(TYPE_ReservationType t)
{
  switch(t)
  {
    case RESERVE_TYPE_Everyday:     return "Everyday";
    case RESERVE_TYPE_EveryWeekend: return "Weekend ";
    case RESERVE_TYPE_Weekly:       return "Weekly  ";
    case RESERVE_TYPE_WeekDay:      return "WeekDay ";
    default:                        return "UNK     ";
  }
}

dword now;
byte weekDay;

dword addTimer(TYPE_ReservationType t)
{
  dword ret, end;
  char *s = ResType(t);
  info.isRec = 1;
  info.reserved = 0xFF; // will be entryNum
  info.tuner = 4; // ANY
  info.svcType = 0; // TV
  info.svcNum = t-1; // 0=BBC1, 2=ITV1
  info.startTime = (now&0xFFFF0000) | 0x173b; // start with today at 23:59, day adjusted depending on the type
  info.reservationType = t;
  switch(t)
  {
    case RESERVE_TYPE_Everyday:     info.startTime -= 21<<16;             break; // Go back a long time
    case RESERVE_TYPE_EveryWeekend: info.startTime -= (21+weekDay+1)<<16; break; // Go back to a Sun (6), subtract weekDay+1
    case RESERVE_TYPE_Weekly:       info.startTime -= 21<<16;             break; // Go back a long time
    case RESERVE_TYPE_WeekDay:      info.startTime -= (21+weekDay+2)<<16; break; // Go back to a Sat (5), subtract weekDay+2
    default: break;
  }
  info.nameFix = 1;
  info.duration = 256; // Triggers timestamp bug
  end = AddTime(info.startTime, info.duration);
  strcpy(info.fileName, "TestTimer1.rec");
  ret = TAP_Timer_Add(&info);
  switch(ret)
  {
    case 0:  TAP_Print(TAPNAME ": %s %d (%4x %02d:%02d - %4x %02d:%02d) timer set\n",           s, info.svcNum, info.startTime>>16, (info.startTime>>8)&0xFF, info.startTime&0xFF, end>>16, (end>>8)&0xFF, end&0xFF); break;
    case -1: TAP_Print(TAPNAME ": %s %d (%4x %02d:%02d - %4x %02d:%02d) no timer slots left\n", s, info.svcNum, info.startTime>>16, (info.startTime>>8)&0xFF, info.startTime&0xFF, end>>16, (end>>8)&0xFF, end&0xFF); break;
    case 1:  TAP_Print(TAPNAME ": %s %d (%4x %02d:%02d - %4x %02d:%02d) invalid or no slot\n",  s, info.svcNum, info.startTime>>16, (info.startTime>>8)&0xFF, info.startTime&0xFF, end>>16, (end>>8)&0xFF, end&0xFF); break;
    case 2:  TAP_Print(TAPNAME ": %s %d (%4x %02d:%02d - %4x %02d:%02d) invalid tuner\n",       s, info.svcNum, info.startTime>>16, (info.startTime>>8)&0xFF, info.startTime&0xFF, end>>16, (end>>8)&0xFF, end&0xFF); break;
    default: TAP_Print(TAPNAME ": %s %d (%4x %02d:%02d - %4x %02d:%02d) clash %8x %2x %2x\n",   s, info.svcNum, info.startTime>>16, (info.startTime>>8)&0xFF, info.startTime&0xFF, end>>16, (end>>8)&0xFF, end&0xFF, ret, (ret>>8)&0xFF, ret&0xFF); break;
  }
  return ret;
}

dword TAP_EventHandler( word event, dword param1, dword param2 )
{
  return param1;
}

int TAP_Main(void)
{
  int count = 0;
  char str[2][80];
  word year;
  byte month, day;
  TYPE_ReservationType type;
  InitTAPex();
  InitTAPAPIFix();
  switch(FixTimerAPI())
  {
    case 0:   TAP_Print(TAPNAME ": bad firmware for TimerAPI patch\n"); break;
    case 1:   TAP_Print(TAPNAME ": TimerAPI patch added\n"); break;
    case 2:   TAP_Print(TAPNAME ": TimerAPI patch already present\n"); break;
    default:  TAP_Print(TAPNAME ": unknown result for TimerAPI patch\n"); break;
  }
  now = Now(NULL);
  TAP_ExtractMjd(now>>16, &year, &month, &day, &weekDay);
  TAP_Print(TAPNAME ": @ %p\n", _tap_startAddr);
  TAP_Print(TAPNAME ": now=%08x %s %2d/%02d/%04d\n",
    now, DaysOfWeek[weekDay], day, month, year);
  str[1][0] = 0;
  for(type = RESERVE_TYPE_EveryWeekend; type < N_ReserveType; type++)
  {
    dword ret = addTimer(type);
    if(ret == 0)
      count++;
    else if(ret == -1)
      strcpy(str[1], "-Ran out of free timers-");
  }
  if(count == 0)
    strcpy(str[0], "Failed to add any timers");
  else if(count<3)
    TAP_SPrint(str[0], "Only added %d timer%s due to clashes", count, count==1?"":"s");
  else
    strcpy(str[0], "Success! Now check the results...");
  ShowMessageWin(TAPNAME, str[0], str[1], 500);
  return 0;
}
